// ::: General utility modules and functions :::
// CC BY 3.0 Tom Øyvind Hogstad <tom.oyvind.hogstad@gmail.com> unless othervise stated by referenced authors

pi = 3.14159265;
F  = 0.001;

// Convenience / Easy read rotate
module rotateX(a) {for(i=[0:$children-1]) {rotate([a,0,0]) child(i);}}
module rotateY(a) {for(i=[0:$children-1]) {rotate([0,a,0]) child(i);}}
module rotateZ(a) {for(i=[0:$children-1]) {rotate([0,0,a]) child(i);}}

// Convenience / Easy read translate
module translateX(a) {for(i=[0:$children-1]) {translate([a,0,0]) child(i);}}
module translateY(a) {for(i=[0:$children-1]) {translate([0,a,0]) child(i);}}
module translateZ(a) {for(i=[0:$children-1]) {translate([0,0,a]) child(i);}}


// Floatingpoint correction for translates
function xF(v) = v - [F,0,0];
function yF(v) = v - [0,F,0];
function zF(v) = v - [0,0,F];

// Vector Readability
function vX(v) = v[0];
function vY(v) = v[1];
function vZ(v) = v[2];

function sqr(a) = pow(a,2);

/* 
Bezier Curve with 3 control points 
Based on bezierconic.scad By Don B, 2011, released into the Public Domain
*/
// Helper function returns single point in sequence
function b3p(p0,p1,p2,steps=5, step=1, i=0) =
	(i==step-1) 
	? (p0+((p1-p0)/steps)*i)+(((p1+((p2-p1)/steps)*i)-(p0+((p1-p0)/steps)*i))*(i/steps)) 
	:b3p(p0=p0,p1=p1, p2=p2, steps=steps, step=step, i= i+1)
;

module bezier3(p0,p1,p2,steps=5, above=false) {
	y_min = p0[1] < p1[1]
		? (p0[1] < p2[1] ? p0[1] : p2[1])
		: (p1[1] < p2[1] ? p1[1] : p2[1]);
	y_max = p0[1] > p1[1]
		? (p0[1] > p2[1] ? p0[1] : p2[1])
		: (p1[1] > p2[1] ? p1[1] : p2[1]);
	echo(y_min, ":", y_max);
	y = (above) ? y_max : y_min;

	for (i=[0:steps-1]) {
		assign(pt1 = b3p(p0=p0,p1=p1,p2=p2,steps=steps, step=i+1) )
		assign(pt2 = b3p(p0=p0,p1=p1,p2=p2,steps=steps, step=i+2) ) {
					
			polygon(points=[pt1,pt2,[pt2[0], y], [pt1[0], y]]);
		}
	}
}


	
// Angle between 2 points
function a2p(p1,p2) = atan2(p2[1]-p1[1],p2[0]-p1[0]);

// Hypotenuse between 2 points
function h2p(p1,p2) = hypotenuse(p2[1]-p1[1],p2[0]-p1[0]);

module bezier3_line(p0,p1,p2,weight=2,steps=5) {
	for (i=[0:steps-1]) {
		assign(pt1 = b3p(p0=p0,p1=p1,p2=p2,steps=steps, step=i+1))
		assign(pt2 = b3p(p0=p0,p1=p1,p2=p2,steps=steps, step=i+2)) 
		assign(pt3 = b3p(p0=p0,p1=p1,p2=p2,steps=steps, step=i+3)) {
			assign(a1 = a2p(pt1, pt2)) 
			assign(a2 = a2p(pt2, pt3))   { 	
				assign(d1 = [weight*sin(a1)*-1, weight*cos(a1)]) 
				assign(d2 = [weight*sin(a2)*-1, weight*cos(a2)]) {
					polygon(points=[pt2, pt2+d1, pt1+d1, pt1]);
					polygon(points=[pt2, pt2+d1, pt3+d2]);

				}
			}
		}
	}
}




p = [[0,18], [12,6], [20,8]];
for(i=p) 
	%translate([i[0], i[1],0]) cylinder(r=1, h=3, $fn=10);
linear_extrude(height=2)
bezier3(p0=p[0], p1=p[1], p2=p[2], weight=2, steps=20, above=true);



// Pythagoras
function hypotenuse(a, b) = sqrt(a*a+b*b);

// Sagitta
function sagitta(r,l) = r - sqrt(r*r-l*l);
//function sagitta_r(s,l) =


// polyR: calculates new radius (R) to get distance between facets to be radius (r)
function poly_r(r, $fn) = r/cos(180/$fn);

// tubeIr: Calculates the inner radius (ir) of a tube with outer radius (or) and wall width (wall), 
// floating point error (fe), slic3r need it to fit exact perimeters
function tubeIr(wall, or, $fn, fe=.01) = (or*cos(180/$fn)-wall-fe)/cos(180/$fn);

module teardrop(h, r, center=true, cut=true, $fn=20)
{
	z  = center ? -h/2 : 0;
	
	translate([0, 0, z])
	difference() {
		union() {
			cylinder(h=h, r=r, $fn=$fn);
			translate([0, r/2*sqrt(2), 0])
			cylinder(h=h, r=r*cos(45), $fn=4);
		}
		if(cut)
		translate([0,r+r*1.6,-z]) cube([r*3, r*3, h+.2], center=center);
	}
}


// http://hydraraptor.blogspot.no/2011/02/polyholes.html
module cylinder_poly(h, r, center=false) {
    n = max(round(4 * r),3);
	echo("cylinder_poly r: ", r, " $fn: ",n);
    cylinder(h = h, r = r / cos (180 / n), $fn=n, center=center);
}


module cylinder_pie(h, r, angle, center=false, $fn=0) {
	dz = center ? -h/2-.1 : -.1;
	difference() {
		cylinder(r=r, h=h, center=center, $fn=$fn);
		rotate(angle/2) translate([0,-r-.1,dz]) cube([r+.1,r*2+.2, h+.2]);
		rotate(-angle/2) translate([-r-.1,-r-.1,dz]) cube([r+.1,r*2+.2, h+.2]);
	}
}

//cylinder_pie(h=20, r=5, angle=10);

module cylinder_tilted(h, r, a, $fn=0) {

	difference() {
		rotateY(a) translate([r, 0,0]) cylinder(r=r, h=h/cos(a)+tan(a)*2*r, $fn=$fn);
		translate([h,0,-h]) cylinder(r=h*2, h=h);
		translate([h,0,h]) cylinder(r=h*2, h=h);

	}
}

//rotate(45) cylinder_tilted(r=5, h=20, a=30);


module tube(h=24, od=15, id=8, split=0, poly=false, center=true, $fn=0) {
	or 	= 	od/2;
	ir	=	id/2;
	z 	= 	center ? 0 : -.01;
	y 	=	center ? -or+(or-ir)/2-.1 :-or-.01 ;
	x 	=	-split/2;
	difference() {
		if (poly) {
			cylinder_poly(h=h, r=or, center=center);
		}
		else {
			cylinder(h=h, r=or, center=center, $fn=$fn);
		}

		if (poly) {
			translate([0,0,z]) cylinder_poly(h=h+.02, r=ir, center=center);
		}
		else {
			translate([0,0,z]) cylinder(h=h+.02, r=ir, center=center, $fn=$fn);
		}
    echo(2*atan((split/2)/(od/2)));
		if (split > 0) {
			translate([0,0,z]) cylinder_pie(h=h+.02, r=od+.01, angle=split, center=center, $fn=$fn);
		}
	}
}

//tube(split=2);

/* Creates a fillet to a cylinder
 r = fillet radius
 h = z height
 dy = fillet y edge above baseline
 rc = cylinder radius
 F  = correction for floatingpoint error
*/
module filletForCylinder(r,h,dy,rc,F=0.02, center=true, $fn=$fn) {
translate([0,0,center ? -h/2 : 0])
difference() {
  translate([(cos(asin((dy+r) / (r+rc)))*rc)-F,dy-F,0]) 
cube([sqrt(pow(r + rc,2) - pow(dy+r,2))-(cos(asin((dy+r) / (r+rc)))*rc)+F,sqrt(pow(rc,2)-pow((cos(asin((dy+r) / (r+rc)))*rc),2))-dy+F,h]);
  translate([sqrt(pow(r + rc,2) - pow(dy+r,2)), r+dy, -F]) cylinder(r=r, h=h+2*F);
}

}



module fillet(r,h,fn=$fn, F=0.01) {
	difference() {
		translate([-F,-F,0]) cube([r+F,r+F,h]);
		translate([r,r,-1]) cylinder(r=r, h=h+2, $fn=fn);
	}
}

